import requests as r
import json

"""Pre-formed HSV for red color"""
redHSV = {'h': 0, 's': 255, 'v': 255}
"""Pre-formed HSV for yellow color"""
yellowHSV = {'h': 60, 's': 255, 'v': 255}
"""Pre-formed HSV for green color"""
greenHSV = {'h': 120, 's': 255, 'v': 255}
"""Pre-formed HSV for sky blue color"""
skyHSV = {'h': 180, 's': 255, 'v': 255}
"""Pre-formed HSV for normal blue color"""
blueHSV = {'h': 240, 's': 255, 'v': 255}
"""Pre-formed HSV for puprle color"""
purpleHSV = {'h': 300, 's': 255, 'v': 255}
"""Pre-formed HSV for white color"""
whiteHSV = {'h': 0, 's': 0, 'v': 255}

"""Pre-formed HSV list of full color spectrum for use with scenes where colors change"""
rainbowHSV = [redHSV, yellowHSV, greenHSV, skyHSV, blueHSV, purpleHSV]

def workModeCommand(wMode = "white"):
    """Generates a command to set the work mode of lights
    
    Returns:
        dict: The formed command to set the work mode of lights

    Args:
        wMode (str, optional): A string corresponding to one of the working modes outlined in the Tuya API documentation. Defaults to "white".
    """
    return {'commands': [{'code': 'work_mode', 'value': wMode}]}

def colorCommand(colorHSV: dict):
    """Generates a command to set the color of lights
    
    Returns:
        dict: The formed command to set the color of lights

    Args:
        colorHSV (dict): The HSV representing desired color

    Raises:
        Exception: Raises exception if color is not in proper HSV format
    """
    if type(colorHSV) is not dict or 'h' not in colorHSV.keys() or 's' not in colorHSV.keys() or 'v' not in colorHSV.keys():
        raise Exception("Colors must be given as HSV in the form {'h': h, 's': s, 'v', v}")
    return {'commands': [{'code': 'colour_data', 'value': colorHSV}]}

def temperatureCommand(newTemperature: int):
    """Generates a command to set the temperature of lights
    
    Returns:
        dict: The formed command to set the color temperature of lights

    Args:
        newTemperature (int): Integer value between 25 and 255, inclusive, representing color temperature

    Raises:
        Exception: Raises exception if newTemperature is outside the range [25,255]
    """
    if newTemperature < 25 or newTemperature > 255:
        raise Exception("Values for temperature must be within the range [25,255], inclusive")
    return {'commands': [{'code': 'temp_value', 'value': newTemperature}]}
    
def brightnessCommand(newBright: int):
    """Generates a command to set the brightness of lights

    Returns:
        dict: The formed command to set the brightness of lights

    Args:
        newBright (int): Integer value between 25 and 255, inclusive, representing brightness

    Raises:
        Exception: Raises exception if newBright is outside the range [25,255]
    """
    if newBright < 25 or newBright > 255:
        raise Exception("Values for brightness must be within the range [25,255], inclusive")
    return {'commands': [{'code': 'bright_value', 'value': newBright}]}

def onCommand():
    """Generates a command to turn on lights
    
    Returns:
        dict: The formed command to turn on lights
    """
    return {'commands': [{'code': 'switch_led', 'value': True}]}

def offCommand():
    """Generates a command to turn off lights

    Returns:
        dict: The formed command to turn off lights
    """
    return {'commands': [{'code': 'switch_led', 'value': False}]}

def sceneCommand(sceneNum = 4, bright = 255, freq = 191, hsvList = rainbowHSV):
    """Generates a command to set the scene mode of lights
    
    Returns:
        dict: The formed command to set the scene

    Args:
        sceneNum (int, optional): The scene number corresponding to desired effects, between 1 and 4. 
                                  See Tuya API documentation for more information on what each scene number means.
                                  Defaults to 4.
        bright (int, optional): Brightness of the lights during the scene. Defaults to 255 (maximum).
        freq (int, optional): Frequency of light changes from scene. Can be between 25 and 255, inclusive. Defaults to 191.
        hsvList (list, optional): List of HSVs representing colors for the scene. Defaults to rainbowHSV.

    """
    cmdCode = "flash_scene_" + str(sceneNum)
    return {'commands': [{
                "code": cmdCode,
                "value": {
                    "bright": bright,
                    "frequency": freq,
                    "hsv": hsvList,
                    "temperature": 0
                }
            }]}


class Light:

    def __init__(self, deviceID: str, deviceName: str):
        """Initialize a new light type device

        Args:
            deviceID (str): The ID of the device, found through the Tuya cloud development interface
            deviceName (str): The custom name you would like to give the light
        """
        self.id = deviceID
        self.name = deviceName
        self.isOn = None
        self.workMode = None
        self.brightness = 0

    def toggleOnOff(self):
        """Generates the right on/off command based on current vals
        Returns:
            dict: Either the command to turn on or to turn off lights 
        """
        if self.isOn:
            return offCommand()
        else:
            return onCommand()

    def postCommand(self, headerData: dict, command: dict):
        """Sends a post request to Tuya API for this device affecting it with the given command

        Args:
            headerData (dict): The data needed for the header to be sent to Tuya API
            command (dict): The command to affect the device with

        Raises:
            Exception: Raises an exception on a failed call to Tuya API
        """
        try:
            thisEnd = "https://openapi.tuyaus.com/v1.0/devices/[id]/commands".replace("[id]", self.id)
            resp = r.post(url=thisEnd, headers=headerData, data=json.dumps(command))
            if resp.json()['success'] == False:
                problem = "ERROR: Command failed, server response => " + json.dumps(resp.json())
                raise Exception(problem)
        except Exception as e:
            raise
        else:
            #print(f"Ran command successfully:\n{command} on device named {self.name}")
            if command['commands'][0]['code'] == "switch_led":
                self.isOn = not self.isOn
            elif command['commands'][0]['code'] == "work_mode" or command['commands'][0]['code'].find("scene") != -1:
                self.workMode = command['commands'][0]['value']
            elif command['commands'][0]['code'] == "bright_value":
                self.brightness = command['commands'][0]['value']

    def setStatusInfo(self, statusInfo: list):
        """Sets the status related variables of this device to the retrieved information from Tuya API. Only used for initializing a device

        Args:
            statusInfo (list): The list contained in the json response to a device status request at the dictionary key 'result'
        """
        self.isOn = statusInfo[0]['value']
        self.workMode = statusInfo[1]['value']
        self.brightness = statusInfo[2]['value']

    def __str__(self):
        return self.name
    
    def __repr__(self):
        return self.name
        


import requests as r
import time
from Crypto.Hash import HMAC, SHA256
import json

class TuyaAPI:

    def __init__(self):
        self.clientID = None
        self.__accessKey = None
        self.__APIHeader = {'client_id': None, 'sign': None, 'sign_method': "HMAC-SHA256", 't': None}
        self.devices = {}

    def __init__(self, id: str, secret: str):
        """Initialize a connection with the Tuya API

        Args:
            id (str): Your Tuya developer account client id
            secret (str): Your Tuya developer account access secret
        """
        self.clientID = id
        self.__accessKey = secret
        self.devices = {}
        try:
            t = str(time.time() * 1000)[:13]
            signature = HMAC.new(str.encode(self.__accessKey), str.encode(self.clientID + t), digestmod=SHA256).hexdigest().upper()
            self.__APIHeader = {'client_id': self.clientID, 'sign': signature, 'sign_method': "HMAC-SHA256", 't': t}
            resp = r.get(url="https://openapi.tuyaus.com/v1.0/token?grant_type=1", headers=self.__APIHeader).json()
            self.tokenTimeLeft = resp['result']['expire_time']
            self.tokenGetTime = time.time()
            self.refreshToken = resp['result']['refresh_token']
            self.__APIHeader['access_token'] = resp['result']['access_token']
            self.refreshSignature()
        except Exception as e:
            raise

    def refreshSignature(self):
        """
        Refreshes the signature in the API head with a new time
        Needs to be done approximately every 10-20 minutes
        """
        t = str(time.time() * 1000)[:13]
        self.__APIHeader['sign'] = HMAC.new(str.encode(self.__accessKey), str.encode(self.clientID + self.__APIHeader['access_token'] + t), digestmod=SHA256).hexdigest().upper()
        self.__APIHeader['t'] = t
        self.signatureGetTime = time.time()

    def refreshAccessToken(self):
        """
        Refresh access token for API
        Must be done every 2 hours to maintain API access
        """
        try:
            t = str(time.time() * 1000)[:13]
            signature = HMAC.new(str.encode(self.__accessKey), str.encode(self.clientID + t), digestmod=SHA256).hexdigest().upper()
            self.__APIHeader = {'client_id': self.clientID, 'sign': signature, 'sign_method': "HMAC-SHA256", 't': t}
            refURL = "https://openapi.tuyaus.com/v1.0/token/" + self.refreshToken
            resp = r.get(url=refURL, headers=self.__APIHeader, params=None).json()
            self.tokenTimeLeft = resp['result']['expire_time']
            self.tokenGetTime = time.time()
            self.refreshToken = resp['result']['refresh_token']
            self.__APIHeader['access_token'] = resp['result']['access_token']
            refreshSignature()
        except Exception as e:
            raise

    def printConvertedTimeLeft(self):
        """Prints the amount of time left until access token must be refreshed"""
        hours = int(self.tokenTimeLeft / 3600)
        minutes = int((self.tokenTimeLeft / 60) - hours * 60)
        seconds = self.tokenTimeLeft % 60
        print(f"Time until key expiry: {hours} Hours, {minutes} Minutes, {seconds} Seconds")
    
    def addDevices(self, devices, identifier: str):
        """Adds devices to the device dictionary and gives them an easier identifier

        Args:
            devices (EasyTuya.devices): A single device object, from one of the devices in EasyTuya.devices, or a list of such objects with the same type.
            identifier (str): The identifier to give the added device(s)
        """
        if identifier not in self.devices.keys():
            self.devices[identifier] = devices
        else:
            if type(devices) is not list:
                self.devices[identifier].append(devices)
            else:
                self.devices[identifier] += devices
        self.__initStatus(identifier)
    
    def sendCommands(self, destIdentifier: str, commands: dict):
        """Send commands to all devices pointed to by destIdentifier

        Args:
            destIdentifier (str): The identifier corresponding to devices added with addDevices()
            commands (dict): The properly formatted commands to send the devices

        Raises:
            Exception: Raises an exception if destIdentifier is not a string
            Exception: Raises an exception if destIdentifier does not exist in devices dictionary
        """
        if time.time() - self.tokenGetTime >= self.tokenTimeLeft:
            self.refreshAccessToken()
        elif time.time() - self.signatureGetTime >= 600:
            self.refreshSignature()
        if type(destIdentifier) is not str:
            raise Exception("ERROR: Command destination identifier must be passed as string")
        if destIdentifier not in self.devices.keys():
            raise Exception("ERROR: Command destination identifier must correspond to a device or group of devices added with addDevice() or addDeviceGroup()")
        try:
            if type(self.devices[destIdentifier]) != list:
                self.devices[destIdentifier].postCommand(self.__APIHeader, commands)
            else:
                for d in self.devices[destIdentifier]:
                    d.postCommand(self.__APIHeader, commands)
        except Exception as e:
            raise

    def __initStatus(self, destID):
        try:
            statuses = self.getStatus(destID)
        except Exception as e:
            raise Exception("Something went wrong while initializing your devices")
        else:
            for d in statuses.keys():
                d.setStatusInfo(statuses[d])

    def getStatus(self, destIdentifier: str):
        """Retrieves the statuses of all devices pointed to by destIdentifier
        returns 

        Args:
            destIdentifier (str): The identifier corresponding to devices added with addDevices()

        Raises:
            Exception: Raises an exception if destIdentifier is not a string
            Exception: Raises an exception if destIdentifier does not exist in devices dictionary

        Returns:
            dict: Device statuses in the form of {deviceObject: results[{}] }
        """
        if time.time() - self.tokenGetTime >= self.tokenTimeLeft:
            self.refreshAccessToken()
        elif time.time() - self.signatureGetTime >= 600:
            self.refreshSignature()
        if type(destIdentifier) is not str:
            raise Exception("ERROR: Status destination identifier must be passed as string")
        if destIdentifier not in self.devices.keys():
            raise Exception("ERROR: Status destination identifier must correspond to a device or group of devices added with addDevice() or addDeviceGroup()")
        try:
            statusURL = "https://openapi.tuyaus.com/v1.0/devices/[id]/status"
            if type(self.devices[destIdentifier]) != list:
                thisURL = statusURL.replace('[id]', self.devices[destIdentifier].id)
                resp = r.get(thisURL, headers=self.__APIHeader).json()
                return {self.devices[destIdentifier]: resp['result']}
            else:
                statusList = {}
                for d in self.devices[destIdentifier]:
                    thisURL = statusURL.replace('[id]', d.id)
                    resp = r.get(thisURL, headers=self.__APIHeader).json()
                    statusList[d] = resp['result']
                return statusList
        except Exception as e:
            raise